#include <cmath>
#include <cstdlib>
#include <ctime>
#include <fstream>
#include <iostream>
#include <string>
using namespace std;

#include "vector.hpp"                 // defines Vector objects
using namespace cpl;                  // in Computational Physics Library

// constant, variable definitions ---------------------------------------------
const double pi = 4 * atan(1.0);      // Math pi
const double g = 9.80;                // acceleration of gravity
const double L = 1.0;                 // pendulum length
const double period = 2 * pi * sqrt(L / g);

double theta0, omega0;                // initial values

// function declarations ------------------------------------------------------

double thetaExact(double);            // exact theta(t)
double omegaExact(double);            // exact omega(t)
double energy(double, double);        // E(theta, omega)

Vector f(Vector);                     // ODE derivative vector function

void EulerStep(                       // take one Euler step
    Vector&,                          // solution vector, input/output
    double);                          // step size

void EulerCromerStep(                 // take one Euler-Cromer step
    Vector&,                          // solution vector, input/output
    double);                          // step size

void RK4Step(Vector&,                 // one 4th order Runge-Kutta step
             double);                 // step size

void adaptiveStep(                    // adaptive step with step doubling
    Vector&,                          // solution vector, input/output
    double&,                          // time step, input/output
    double,                           // desired accuracy
    void method(Vector&, double)      // algorithm to use
        = RK4Step);                   // default if argument omitted

void simulate(
    Vector&,                          // trajectory point, input/output
    int,                              // number of periods
    double&,                          // time step
    double,                           // desired accuracy
    string method);                   // integration method

// main program ---------------------------------------------------------------

int main() {
    cout << "Simulation of Simple Pendulum using adaptive integration\n"
         << "--------------------------------------------------------\n";
    cout << "Enter number of periods: ";
    int periods;
    cin >> periods;    
    cout << "Enter initial theta and omega: ";
    cin >> theta0 >> omega0;
    cout << "Enter accuracy: ";
    double accuracy;
    cin >> accuracy;

    Vector x0(3), x(3);
    x0[0] = 0;
    x0[1] = theta0;
    x0[2] = omega0;

    double dt = 0.001;
    simulate(x = x0, periods, dt, accuracy, "Euler");
    dt = 0.001;
    simulate(x = x0, periods, dt, accuracy, "Euler-Cromer");
    dt = 0.001;
    simulate(x = x0, periods, dt, accuracy, "Runge-Kutta");
}

// function definitions -------------------------------------------------------

double relativeError(double a, double b) {
   double scale = 0.5 * abs(a + b);
   if (scale == 0.0) 
       return 0.0;
   return abs(a - b) / scale;
}

// Simulation function
void simulate(Vector& x, int periods, double& dt, 
              double accuracy, string method) 
{
    double dtMin = dt = 0.1;
    Vector thetaErrorAv(periods), omegaErrorAv(periods);
    Vector energyErrorAv(periods);
    int totalSteps = 0;

    cout << "Using " << method << " ..." << endl;
    clock_t t1 = clock();

    for (int p = 0; p < periods; ++p) {
        int steps = 0;
        double thetaError = 0, omegaError= 0, energyError = 0;

        while (x[0] < (p + 1) * period) {
            if (method == "Euler")
                adaptiveStep(x, dt, accuracy, EulerStep);
            else if (method == "Euler-Cromer")
                adaptiveStep(x, dt, accuracy, EulerCromerStep);
            else
                adaptiveStep(x, dt, accuracy);
            if (dt < dtMin)
                dtMin = dt;
            thetaError  += abs(x[1] - thetaExact(x[0]));
            omegaError  += abs(x[2] - omegaExact(x[0]));
            energyError += abs(energy(x[1], x[2]) - energy(theta0, omega0));
            ++steps;
        }

        thetaErrorAv[p] = thetaError / steps;
        omegaErrorAv[p] = omegaError / steps;
        energyErrorAv[p] = energyError / steps;

        totalSteps += steps;
    }

    clock_t t2 = clock();

    cout << "Smallest step size = " << dtMin << '\n'
         << "Total number of steps = " << totalSteps << '\n'
         << "CPU time = " << (t2 - t1) / double(CLOCKS_PER_SEC) << " seconds" 
         << endl;

    ofstream errorFile(method.c_str());
    for (int p = 0; p < periods; p++)
        errorFile << (p + 1.0) << '\t'
                  << thetaErrorAv[p] << '\t' 
                  << omegaErrorAv[p] << '\t' 
                  << energyErrorAv[p] << '\n';

    cout << "Errors written in file " << method << endl;
    errorFile.close();
}

// Exact solutions

double thetaExact(double t) {
    return theta0 * cos(2 * pi * t / period) +
           period / (2 * pi) * omega0 * sin(2 * pi * t / period);
}

double omegaExact(double t) {
    return omega0 * cos(2 * pi * t / period) -
           2 * pi / period * theta0 * sin(2 * pi * t / period);
}

// Energy
double energy(double theta, double omega) {
    return 0.5 * (omega * omega + (g / L) * theta * theta);
}

// Derivative vector function
Vector f(Vector x) {
    double t = x[0], theta = x[1], omega = x[2];
    Vector f(3);
    f[0] = 1;
    f[1] = omega;
    f[2] = - (g/L) * theta;
    return f;
}

//  Euler
void EulerStep(Vector& x, double tau) {
    x += tau * f(x);
}

//  Euler-Cromer is NOT a vector algorithm
void EulerCromerStep(Vector& x, double dt) {
    double t = x[0], theta = x[1], omega = x[2];
    omega -= (g/L) * theta * dt;
    theta += omega * dt;
    t += dt;
    x[0] = t, x[1] = theta, x[2] = omega;
}

//  Fourth order Runge-Kutta
void RK4Step(Vector& x, double tau) {
    Vector k1 = tau * f(x);
    Vector k2 = tau * f(x + 0.5 * k1);
    Vector k3 = tau * f(x + 0.5 * k2);
    Vector k4 = tau * f(x + k3);
    x += (k1 + 2 * k2 + 2 * k3 + k4) / 6.0;
}

//  Adaptive step size control using step doubling
//  adapted from Numerical Recipes to apply also to Euler
void adaptiveStep(Vector& x, double& tau, double accuracy,
                  void method(Vector&, double)) 
{
    const double SAFETY = 0.9, TINY = 1.0e-30;
    double p = 5, pGrow, pShrink, errCon, gravy = 1 / 15.0;
    if (method == EulerStep || method == EulerCromerStep) 
        p = 2, gravy = 0;
    pGrow = - 1 / p;
    pShrink = - 1 / (p - 1);
    errCon = pow(5 / SAFETY, 1 / pGrow);
    int n = x.size();
    Vector x_half(n), x_full(n), Delta(n);
    Vector scale = f(x);
    for (int i = 0; i < n; i++)
        scale[i] = abs(x[i]) + abs(scale[i] * tau) + TINY;
    double err_max;
    while (true) {
        // take two half steps
        double tau_half = tau / 2;
        x_half = x;
        method(x_half, tau_half);
        method(x_half, tau_half);
        // take full step
        x_full = x;
        method(x_full, tau);
        // estimate error
        Delta = x_half - x_full;
        err_max = 0;
        for (int i = 0; i < n; i++)
            err_max = max(err_max, abs(Delta[i]) / scale[i]);
        err_max /= accuracy;
        if (err_max <= 1.0)
            break;
        double tau_temp = SAFETY * tau * pow(err_max, pShrink);
        if (tau >= 0.0)
            tau = max(tau_temp, 0.1 * tau);
        else
            tau = min(tau_temp, 0.1 * tau);
        if (abs(tau) == 0.0) {
            cerr << "adaptiveStep: step size underflow, aborting ..." << endl;
            exit(EXIT_FAILURE);
        }
    }
    tau *= (err_max > errCon ? SAFETY * pow(err_max, pGrow) : 5.0);
    x = x_half + Delta * gravy;
}

